from entry.api.exceptions import ApiNotAuthenticated, NoApiTokenFound, PermissionScopeDenied
from urllib.parse import parse_qs
from masonite.routes import Post, Delete
from entry.api.models.OAuthToken import OAuthToken


class TokenAuthentication:

    authentication_model = OAuthToken
    scopes = ['*']

    def authenticate(self):

        # Find which input has the authorization token:
        if self.request.has('token'):
            token = self.request.input('token')
        elif self.request.header('AUTHORIZATION'):
            token = self.request.header('AUTHORIZATION').replace('Bearer ', '')

        if not self.authentication_model:
            self.authentication_model = self.model

        # Check if Authentication token exists
        if not self.request.header('AUTHORIZATION') and not self.request.has('token'):
            raise NoApiTokenFound

        if not self.authentication_model.where('token', token).count():
            raise ApiNotAuthenticated
        
        # Check correct scopes:
        if '*' not in self.scopes:
            scopes = self.authentication_model.where('token', token).first().scope.split(' ')
            
            if not set(self.scopes).issubset(scopes):
                raise PermissionScopeDenied

        # Delete the token input
        if self.request.has('token') and self.request.is_not_get_request():
            build_new_inputs = {}
            for i in self.request.params:
                build_new_inputs[i] = self.request.params[i]
            
            build_new_inputs.pop('token')
            self.request.params = build_new_inputs
    
        
    def tokens_from_model(self, model):
        self.authentication_model = model
        return self
    

    @staticmethod
    def routes():
        try:
            return [
                Post().module('app.http.controllers.Entry.Api').route('/oauth/token', 'OAuthPasswordGrantController@generate'),
                Delete().module('app.http.controllers.Entry.Api').route('/oauth/token', 'OAuthPasswordGrantController@revoke'),
            ]
        except ImportError as e:
            print("\033[93mWarning: could not find app.http.controllers.Entry.Api - Error {0}".format(e))
        
        return []



import os
import subprocess
from cleo import Command
from masonite.packages import create_controller, append_web_routes, append_api_routes


package_directory = os.path.dirname(os.path.realpath(__file__))

class InstallCommand(Command):
    """
    Installs needed controllers and routes into a Masonite project

    entry:install
    """

    def handle(self):
        
        create_controller(
            os.path.join(package_directory,
                         '../entry_snippets/controllers/OAuthPasswordGrantController.py'),
            to='app/http/controllers/Entry/Api'
        )

        append_web_routes(
            os.path.join(package_directory,
                         '../entry_snippets/routes/EntryRoutes.py')
        )

        subprocess.call(['craft', 'migrate'])



from entry.api.exceptions import RateLimitReached

class RateLimit:
    rate_limit = (100, 1, 'minute')

    def limit(self):
        cache = self.container.make('Cache')
        if not cache.is_valid('Entry/rate_limit/IP/' + self.request.environ['REMOTE_ADDR']):
            # create a new cache
            cache.store_for(
                'Entry/rate_limit/IP/' +
                self.request.environ['REMOTE_ADDR'], '1',
                self.rate_limit[1], self.rate_limit[2]
            )
            self.request.header('X-RateLimit-Limit', str(self.rate_limit[0]), http_prefix = None)
            self.request.header('X-RateLimit-Remaining', str(int(self.rate_limit[0]) - 1), http_prefix = None)
            return
        else:
            rate_limit = int(cache.get(
                'Entry/rate_limit/IP/' + self.request.environ['REMOTE_ADDR']
            )) + 1

            if rate_limit > self.rate_limit[0]:
                raise RateLimitReached

            cache.update(
                'Entry/rate_limit/IP/' + self.request.environ['REMOTE_ADDR'],
                rate_limit
            )
            # Set Headers
            self.request.header('X-RateLimit-Limit', str(self.rate_limit[0]), http_prefix = None)
            self.request.header('X-RateLimit-Remaining', str(int(self.rate_limit[0]) - rate_limit), http_prefix = None)
            return
        




from orator.migrations import Migration


class AddTokenAuthentication(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create('oauth_tokens') as table:
            table.increments('id')
            table.integer('user_id').unsigned()
            table.foreign('user_id').references('id').on('users')
            table.string('name')
            table.string('scope')
            table.string('token')
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop('oauth_tokens')



""" An ApiProvider Service Provider """
import os

from entry.commands.InstallCommand import InstallCommand
from entry.commands.ResourceCommand import ResourceCommand
from masonite.packages import create_controller
from masonite.provider import ServiceProvider
from routes import api

package_directory = os.path.dirname(os.path.realpath(__file__))

class ApiProvider(ServiceProvider):

    def register(self):
        self.app.bind('ResourceRoutes', api.RESOURCES)
        self.app.bind('InstallCommand', InstallCommand())
        self.app.bind('ResourceCommand', ResourceCommand())
        self.app.bind(
            'Entry_MigrationDirectory',
            os.path.join(package_directory, '../migrations')
        )


    def boot(self, Response, ResourceRoutes, Route, Request, Headers):
        router = Route
        if Response == 'Route not found. Error 404':

            """
            |--------------------------------------------------------------------------
            | Pull in the API Routes from the Service Container
            |--------------------------------------------------------------------------
            |
            | The Service Container has loaded all Api routes into the container so
            | let's loop through and check for any matches.
            |
            """

            for route in ResourceRoutes:

                """
                |--------------------------------------------------------------------------
                | If We've Got A Match
                |--------------------------------------------------------------------------
                |
                | If we have a match then let's go ahead and execute the route, load the 
                | response into the data variable and get on with our lives.
                |
                """

                if route.url in router.url:
                    data = route.load_request(Request).load_container(self.app).handle()

                    if data:
                        self.app.bind('Response', data)
                        Headers += [
                            ("Content-Type", "application/json; charset=utf-8")
                        ]
                        break
                    else:
                        data = 'Route not found. Error 404'
                else:
                    data = 'Route not found. Error 404'



''' A Module Description '''
from app.User import User
from masonite.facades.Auth import Auth
from entry.api.models.OAuthToken import OAuthToken


class OAuthPasswordGrantController:
    ''' Class Docstring Description '''

def generate(self, Request):
    if not Request.has('username') or not Request.has('password'):
        return {'error': 'This API call requires a username and password in the payload.'}

    user = Auth(Request).login(Request.input(
        'username'), Request.input('password'))

    if user:
        if Request.has('scopes'):
            scopes = Request.input('scopes')
        else:
            scopes = ''

        return {'token': user.create_token(scopes=scopes)}
    else:
        return {'error': 'Incorrect username or password'}

    def revoke(self, Request):
        if not Request.has('token'):
            return {'error': 'Token not received'}

        get_token = OAuthToken.where('token', Request.input('token')).first()
        if get_token:
            get_token.delete()
            return {'success': 'Token was revoked'}
        else:
            return {'error': 'Could not find token'}



import uuid
from entry.api.models.OAuthToken import OAuthToken

class HasApiTokens:
    """
        Helper Class For API Authentication Models
    """


    def prevent_scopes(self):
        pass


    def grant_scopes(self):
        pass


    def create_token(self, name='default', scopes=''):
        """ Create a new API token and save to model """

        if isinstance(self.grant_scopes(), list):
            scopes = self.grant_scopes()

        if isinstance(scopes, list):
            if isinstance(self.prevent_scopes(), list):
                scopes = set(scopes) - set(self.prevent_scopes())
            scopes = ' '.join(scopes)
        else:
            if isinstance(self.prevent_scopes(), list):
                scopes = scopes.split(' ')
                scopes = set(scopes) - set(self.prevent_scopes())
                scopes = ' '.join(scopes)

        generated_token = uuid.uuid4().hex

        user_by_token = OAuthToken.where('user_id', self.id).first()

        if user_by_token:
            # update the token
            user_by_token.token = generated_token
            user_by_token.scope = scopes
            user_by_token.save()
        else:
            OAuthToken.create(
                user_id = self.id,
                name = name,
                scope = scopes,
                token = generated_token
            )

        return generated_token


    def get_token(self):
        """ Get current API Token """
        return OAuthToken.where('user_id', self.id).first().token


    def has_token(self):
        """ Check if user has an API token """

        if OAuthToken.where('user_id', self.id).first():
            return True
        
        return False
    
    def has_scope(self, scope):
        get_token = OAuthToken.where('user_id', self.id).first()
        
        if get_token:
            scopes = get_token.scope.split(' ')
            if scope in scopes:
                return True
        return False
    
    def add_scopes(self, *scopes):
        get_token = OAuthToken.where('user_id', self.id).first()
        
        if get_token:
            scopes = ' '.join(scopes)
            get_token.scope += scopes
            return get_token.save()

        return False
    

    def set_scopes(self, *scopes):
        get_token = OAuthToken.where('user_id', self.id).first()
        get_token.scope = ''
        get_token.save()
        return self.add_scopes(scopes)


    def with_token(self, token):
        """ Set API token for the current user """
        self.token = token
        return self
    
    def revoke_token(self):
        get_token = OAuthToken.where('user_id', self.id).first()
        if get_token:
            return get_token.delete()
        
        return False



from setuptools import setup

setup(
    name="masonite-entry",
    version='0.1.5',
    packages=[
        'entry',
        'entry.api',
        'entry.api.models',
        'entry.commands',
        'entry.migrations',
        'entry.providers',
    ],
    install_requires=[
        'masonite>=1.5,<= 1.6.99',
    ],
    include_package_data=True,
)



import json


class JsonSerialize:
    
    def serialize(self, data):
        if isinstance(data, dict):
            if self.data_wrap:
                return json.dumps({'data': data})
            
            return json.dumps(data)
        else:
            if self.data_wrap:
                return json.dumps({'data': json.loads(data.to_json())})
            
            return data.to_json()
        




import json
import bcrypt
import re
from entry.api.exceptions import (
    ApiNotAuthenticated,
    NoApiTokenFound,
    PermissionScopeDenied,
    RateLimitReached,
)

class Resource:

    exclude = []
    exclude_relationship_fields = []
    methods = ['create', 'read', 'update', 'delete']
    model = None
    relationships = None
    request = None
    url = None
    url_prefix = ''
    read_only_fields = []
    data_wrap = True


    def __init__(self):
        self.model.__hidden__ = self.exclude
        if not self.url:
            self.url = '/api/{0}'.format(self.model().__class__.__name__.lower())
        
        if self.url_prefix:
            self.url = '{0}{1}'.format(self.url_prefix, self.url)

        self.container = None

    def handle(self):
        # Run authentication if one exists
        if hasattr(self, 'authenticate'):
            try:
                self.authenticate()
            except ApiNotAuthenticated:
                self.request.status('401 Unauthorized')
                return json.dumps({'error': 'Invalid authentication token'})
            except NoApiTokenFound:
                self.request.status('400 Bad Request')
                return json.dumps({'error': 'Authentication token not found'})
            except PermissionScopeDenied:
                self.request.status('401 Unauthorized')
                return json.dumps({'error': 'Incorrect permission scope'})

        # Run rate limiting if one exists
        if hasattr(self, 'limit'):
            try:
                self.limit()
            except RateLimitReached:
                # Set Headers
                self.request.header('X-RateLimit-Limit', str(self.rate_limit[0]), http_prefix=None)
                self.request.header('X-RateLimit-Remaining', '0', http_prefix=None)
                self.request.status('429 Too Many Requests')
                return json.dumps({'error': 'Rate limit of {0} calls every {1} {2} reached'.format(self.rate_limit[0], self.rate_limit[1], self.rate_limit[2])})

        self.request.header('Allowed', ', '.join(self._get_http_verbs()), http_prefix=None)

        for item in list(self.model().__dict__):
            if item.startswith('_'):
                self.model().__dict__.pop(item)

        try:
            if self.request.environ['REQUEST_METHOD'] == 'POST' and 'create' in self.methods:
                self.request.status('201 Created')
                return self.serialize(self.create())

            if self.request.environ['REQUEST_METHOD'] == 'GET' and 'read' in self.methods:
                self.request.status('200 OK')
                return self.serialize(self.read())

            if self.request.environ['REQUEST_METHOD'] == 'PUT' and 'update' in self.methods:
                self.request.status('200 OK')
                return self.serialize(self.update())

            if self.request.environ['REQUEST_METHOD'] == 'DELETE' and 'delete' in self.methods:
                self.request.status('200 OK')
                return self.serialize(self.delete())
        except Exception as e:
            self.request.status('500 Internal Server Error')
            return self.serialize({'error': str(e)})

        self.request.status('405 Method Not Allowed')
        return self.serialize(
            {
                'Error': 'Invalid URI: {0} {1}. This route does not exist for this endpoint.'.format(
                    self.request.environ['REQUEST_METHOD'], self.request.path)
            }
        )

    
    def load_request(self, request):
        self.request = request
        return self

    def load_container(self, container):
        self.container = container
        return self

    def create(self):
        if '{0}s'.format(self.url) == self.request.path:
            # if POST /api/users
            proxy = self.model()
            for field in self.request.all():
                # If the field is a password, hash it
                if field == 'password':
                    password = bcrypt.hashpw(
                        bytes(self.request.input('password'),
                              'utf-8'), bcrypt.gensalt()
                    )
                    setattr(proxy, field, password)
                else:
                    setattr(proxy, field, self.request.input(field))
            proxy.save()

            return proxy

    def read(self):
        matchregex = re.compile(r'^[\/\w+]+\/(\d+)')
        match_url = matchregex.match(self.request.path)

        # Get the plural of the url
        if '{0}s'.format(self.url) == self.request.path:
            # if GET /api/users
            models = self.model().all()
            if self.relationships:
                for model in models:
                    self._get_relationships(model)

            return models
        elif match_url:
            record = self.model().find(match_url.group(1))
            if record:
                if self.relationships:
                    self._get_relationships(record)
                return record
            else:
                return {'Error': 'Record Not Found'}

        return {'Error': 'Invalid URI: {0}'.format(self.request.path)}

    def update(self):
        # if PUT /api/user/1
        matchregex = re.compile(r"^\/\w+\/\w+\/(\d+)")
        match_url = matchregex.match(self.request.path)

        proxy = self.model.find(match_url.group(1))
        if not proxy:
            return {'Error': 'Record Not Found'}

        for field in self.request.all():
            if field not in self.read_only_fields:
                setattr(proxy, field, self.request.input(field))
        proxy.save()
        proxy = self.model.find(match_url.group(1))
        return proxy

    def delete(self):
        # if DELETE /api/user/1
        matchregex = re.compile(r"^\/\w+\/\w+\/(\d+)")
        match_url = matchregex.match(self.request.path)

        get = self.model.find(match_url.group(1))
        if get:
            get.delete()
            return get
        else:
            return {'Error': 'Record Not Found'}

    def _get_relationships(self, model):
        # Get relationships
        for relationship in self.relationships:
            if hasattr(model, relationship):
                attr = getattr(model, relationship)
                if attr and relationship in self.exclude_relationship_fields:
                    attr.__hidden__ = self.exclude_relationship_fields[relationship]

    def _get_http_verbs(self):
        verbs = []

        if 'create' in self.methods:
            verbs.append('POST')
        if 'read' in self.methods:
            verbs.append('GET')
        if 'update' in self.methods:
            verbs.append('PUT')
        if 'delete' in self.methods:
            verbs.append('DELETE')
        
        return verbs



import os
from cleo import Command


package_directory = os.path.dirname(os.path.realpath(__file__))


class ResourceCommand(Command):
    """
    Creates a new API resource

    entry:resource
        {name : Name of the resource you want to create}
    """

    def handle(self):
        resource = self.argument('name')
        if not os.path.isfile('app/http/resources/{0}.py'.format(resource)):
            if not os.path.exists(os.path.dirname('app/http/resources/{0}.py'.format(resource))):
                # Create the path to the resource if it does not exist
                os.makedirs(os.path.dirname('app/http/resources/{0}.py'.format(resource)))

            f = open('app/http/resources/{0}.py'.format(resource), 'w+')

            f.write("''' A {0} API Resource '''\n".format(resource))
            f.write('from entry.api.Resource import Resource\n')
            f.write('from entry.api.JsonSerialize import JsonSerialize\n\n\n')
            f.write("class {0}(Resource, JsonSerialize):\n    model = None\n".format(resource))
            f.close()
            self.info('Resource Created Successfully!')
        else:
            self.error('Resource Already Exists!')
